import regex 
from typing import Union

import chompjs

regex.Default_VERSION = regex.VERSION0

def extract_js_content(content: Union[list, str]) -> dict[str, Union[dict, list, str]]:
    """
    Extracts from scraped JavaScript content, which is either a list of strings
    or a single string, all dictionaries and lists and parses them
    to Python lists or dictionaries.

    The result is a dictionary containing the name of the var as the key and 
    the parsed content as value.


    Args:
        content (Union[str, list[str]]): List of strings or a single string that
                                         contain JavaScript code.

    Raises:
        ValueError: If there is already a parsed object in the result dictionary with
                    the identical name. Might be handeld in a later version.

    Returns:
        dict[str, Union[dict, list]]: Dict containing the name of the var as the key and
                                      its content as value.
    """

    js_var_exp: str = r'(?:_|\$|\p{L})+[\$_\w]*'

    single_val_exp: str = fr'\s*(?:var|let|const)\s+{js_var_exp}\s*=\s*(?:".*?"|\d+(?:\.\d)*)\s*'
    dict_exp: str = fr'((?:var|let|const)\s+{js_var_exp}\s*=\s*(\{{[^\\}}\{{]*+(?:(?2)[^\}}\{{]*)*+\}}))'
    list_exp: str = fr'(\s*{js_var_exp}\s*=+\s*(\[[^\]\[]*+(?:(?2)[^\]\[]*)*+\]))'

    
    if type(content) == str:
        content = [content]

    js_objects: list[str] = []

    content = [item.replace('\n', ' ') for item in content]
    
    for item in content:
        single_values: list[str] = regex.findall(single_val_exp, item, regex.DOTALL)
        for single_value in single_values:
            eq_pos = single_value.find('=')+1
            js_objects.append(single_value[:eq_pos] + '[' + single_value[eq_pos:] + ']') #so its valid json

        for expression in [list_exp, dict_exp]:
            js_objects.extend([result[0] for result in regex.findall(expression, item, regex.DOTALL)])

        
    results: dict[str, Union[dict, list]] = {}
    for obj in js_objects:
        parsed_obj = chompjs.parse_js_object(obj)

        declar_split = obj.split('=')[0].lstrip()
        if declar_split[:4] == 'var ': #note the space so strings like "variable" are false
            obj_name = declar_split.split(' ')[1].strip()
        else:
            obj_name = declar_split.strip()

        if obj_name in results.keys():
            #todo: what to do here?
            raise ValueError('Object Name already parsed.')
        else:
            results[obj_name] = parsed_obj

    return results


from distutils.core import setup
setup(
  name = 'js_extractor',         
  packages = ['js_extractor'],   
  version = '0.1',      
  license='MIT',        
  description = 'Find JavaScript Variables, Dicts and Lists in a String and parse them with their content into a Python Dictionary',   # Give a short description about your library
  author = 'Bjarne Schroeder',                   
  author_email = 'schroeder.bjarne97@gmail.com',      
  url = 'https://github.com/bjarneschroeder/js_extractor',   
  download_url = 'https://github.com/bjarneschroeder/js_extractor/archive/refs/tags/v_01.tar.gz',    # I explain this later on
  keywords = ['scraping', 'parsing', 'python', 'extracting', 'javascript'],   # Keywords that define your package best
  install_requires=[
        'atomicwrites',
        'attrs',
        'chompjs',
        'colorama',
        'iniconfig',
        'packaging',
        'pluggy',
        'py',
        'pyparsing',
        'pytest',
        'regex',
        'toml'
  ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Libraries :: Python Modules',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],
)


from typing import Union
import pytest
from src.js_extractor import extract_js_content


@pytest.mark.parametrize('content, expected_result', [
    ('var content = "value1"',                      {'content': ['value1']}),
    ('var content="value1";',                       {'content': ['value1']}),
    ('asdf var content = "value1"\\n sde1sfd',      {'content': ['value1']}),
    ('var content = "value1"',                      {'content': ['value1']}),
    ('var content = "8.0";',                          {'content': ['8.0']}),
    ('asdf var content = "8.0" ;asdf',                {'content': ['8.0']}),
])
def test_single_value_extraction(content: str, expected_result: dict[str, str]):
    assert extract_js_content(content) == expected_result


@pytest.mark.parametrize('content, expected_result', [
    ('var content = {"key1": "value1"}', {'content': {'key1': 'value1'}}),
    ('var CONTENT = {"key1": "value1"}', {'CONTENT': {'key1': 'value1'}}),
    ('var c0nt3nt = {"key1": "value1"}', {'c0nt3nt': {'key1': 'value1'}}),
    ('var con_tent = {"key_1": "value_1"}', {'con_tent': {'key_1': 'value_1'}}),
    ('\nvar content = {"key1": "value1"}',   {'content': {'key1': 'value1'}})
])
def test_all_name_kinds(content: str, expected_result: dict[str, str]):
    assert extract_js_content(content) == expected_result

@pytest.mark.parametrize('content',[
    ('var con-tent= {"key1": "value1"}'),
    ('let 1content = {"key1": "value1"}'),
    ('asdfvar = {"key1": "value1"}')
])
def test_false_positive_names(content: str):
    assert extract_js_content(content) == {}

@pytest.mark.parametrize('content, expected_result', [
    ('var content = {"key1": "value1"}',            {'content': {'key1': 'value1'}}),
    ('var content = {"key1": "value1"};',           {'content': {'key1': 'value1'}}),
    ('var content={"key1":"value1"}',               {'content': {'key1': 'value1'}}),
    ('var content={"key1":"value1"};',              {'content': {'key1': 'value1'}}),
    ('adsf var content = {"key1": "value1"};asdf',  {'content': {'key1': 'value1'}})
])
def test_dict_extraction(content: str, expected_result: dict[str, str]):
    assert extract_js_content(content) == expected_result

@pytest.mark.parametrize('content, expected_result', [
    ('var content = {"nest1": {"n1_key1": "n1_value1"}}',                                                   {'content': {'nest1': {'n1_key1': 'n1_value1'}}}),
    ('var content = {"nest1": {"n1_key1": "n1_value1"}, "key1": "value1"}',                                 {'content': {'nest1': {'n1_key1': 'n1_value1'}, 'key1': 'value1'}}),
    ('var content = {"nest1": {"n1_key1": "n1_value1", "n1_key2": "n1_value2"}}',                           {'content': {'nest1': {'n1_key1': 'n1_value1', 'n1_key2': 'n1_value2'}}}),
    ('var content = {"nest1": {"n1_key1": "n1_value1", "n1_key2": "n1_value2"}, "key1": "value1"}',         {'content': {'nest1': {'n1_key1': 'n1_value1', 'n1_key2': 'n1_value2'}, 'key1': 'value1'}}),
    ('var content = {"nest1": {"n1_nest1": {"n1_n1_key1": "n1_n1_value1"}}}',                               {'content': {'nest1': {'n1_nest1': {'n1_n1_key1': 'n1_n1_value1'}}}}),
    ('var content = {"nest1": {"n1_nest1": {"n1_n1_nest1": {"n1_n1_n1_key1": "n1_n1_n1_value1"}}}, "nest2": {"n2_nest1":{"n2_n1_key1": "n2_n2_value1"}}}', 
        {'content': {'nest1': {'n1_nest1': {'n1_n1_nest1': {'n1_n1_n1_key1': 'n1_n1_n1_value1'}}}, 'nest2': {'n2_nest1': {'n2_n1_key1': 'n2_n2_value1'}}}})
])
def test_nested_dict_extraction(content, expected_result: dict[str, str]):
    assert extract_js_content(content) == expected_result


@pytest.mark.parametrize('content, expected_result', [
    ('content = ["ele1", "ele2"]',              {'content': ['ele1', 'ele2']}),
    ('content = ["ele1", "ele2"];',             {'content': ['ele1', 'ele2']}),
    ('content=["ele1", "ele2"]',                {'content': ['ele1', 'ele2']}),
    ('content=["ele1", "ele2"];',               {'content': ['ele1', 'ele2']}),
    ('asdfe content=["ele1", "ele2"];asdfasf',  {'content': ['ele1', 'ele2']}),
])
def test_list_extraction(content: str, expected_result: dict[str, list[str]]):
    assert extract_js_content(content) == expected_result


@pytest.mark.parametrize('content, expected_result', [
    ('content = [["n1_ele1", "n1_ele2"], "ele1"]',                          {'content': [['n1_ele1', 'n1_ele2'], 'ele1']}),
    ('content = [{"n1_key1": "n1_value1"}, "ele1"]',                        {'content': [{'n1_key1': 'n1_value1'}, 'ele1']}),
    ('content = [{"n1_key1": ["n1_ele1", "n1_ele2"]}, ["ele1", "ele2"]]',   {'content': [{'n1_key1': ['n1_ele1', 'n1_ele2']}, ['ele1', 'ele2']]}),
])
def test_nested_list_extraction(content: str, expected_result: list[Union[dict[list[str]], list[str]]]):
    assert extract_js_content(content) == expected_result


@pytest.mark.usefixtures('content_list')
def test_multiple_js_contents(content_list):
    assert extract_js_content(content_list['content']) == content_list['result']

@pytest.mark.usefixtures('content_list')
def test_multiple_parameters(content_list):
    contents_as_list: list[str] = content_list['content']
    content: str = '\n'.join(contents_as_list)
    assert  extract_js_content(content) == content_list['result']